﻿namespace HeuristicLab.BenchmarkSuite.Problems {
  using System;
  using System.Linq;

  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc.Char;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc.Integer;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc.IntegerVector;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc.String;

  public class WordStats : BenchmarkSuiteDataDescriptor {
    private const string name = "Word Stats - Hard";
    private const string fileName = "WordStats.csv";
    private const string description = "Given a string, print the number of words containing n characters for n from 1 to the length of the longest word. At the end of the output, print a line that gives the number of sentences and line that gives the average sentence length. A word is any string of consecutive non-whitespace characters(including sentence terminators). Every string will contain at least one sentence terminator(period, exclamation point, or question mark). The average sentence length is the number of words in the file divided by the number of sentence terminator characters.";

    protected override string FileName { get { return fileName; } }
    public override string Name { get { return name; } }
    public override string Description { get { return description; } }
    protected override int InputArgumentCount { get { return 1; } }
    protected override int OutputArgumentCount { get { return 1; } }

    public override ProblemData CreateProblemData() {
      return new ProblemData(ProblemType.WordStats) {
        Name = Name,
        Description = Description,
        ProgramExecutionBudget = 30000000,
        Examples = CloneExamples(),
        BestResult = 0,
        WorstResult = 10000,
        InputArgumentTypes = new[] { ExampleArgumentType.String },
        OutputArgumentTypes = new[] { ExampleArgumentType.Print },
        TrainingCount = 100,
        TestCount = 1000,
        EnabledDataTypes = DataTypes.Exec | DataTypes.Integer | DataTypes.Float | DataTypes.Boolean | DataTypes.Char | DataTypes.String | DataTypes.IntegerVector | DataTypes.FloatVector | DataTypes.StringVector | DataTypes.Print,
        MaxSize = 1000,
        EvalLimit = 6000,
        FloatStringFormat = "0.0#########",
        ErcOptions = {
        ErcProbability = 0.05,
          IntegerErcOptions = new IntegerErcOptions(
            new IntegerRangeErc(-100, 100)),
          IntegerVectorErcOptions = new IntegerVectorErcOptions(
            new IntegerVectorConstantsErc(new int[0])),
          CharErcOptions = new CharErcOptions(
            new IntegerConstantErc('.', '?', '!', ' ', '\t', '\n', ':')),
          StringErcOptions = new StringErcOptions(
            new StringConstantErc("words of length ", ": ", "number of sentences: ", "average sentence length: "))
        }
      };
    }

    protected override Example ParseExample(string[] input, string[] output) {
      long numberOfSentences;
      if (!GetNumberOfSentences(output[0], out numberOfSentences))
        throw new InvalidOperationException("Unable to find or parse line with number of sentences.");

      double averageSentenceLength;
      if (!GetAverageSentenceLength(output[0], out averageSentenceLength))
        throw new InvalidOperationException("Unable to find line or parse with average sentence length.");

      return new Example {
        InputArgs = input,
        OutputArgs = output,
        InputString = input,
        OutputPrint = output[0],

        // help
        OutputInteger = new[] { numberOfSentences },
        OutputFloat = new[] { averageSentenceLength }
      };
    }

    public static bool GetNumberOfSentences(string str, out long numberOfSentences) {
      var lines = str.Split('\n');
      var requiredLineIndex = lines.Length - 2;
      numberOfSentences = 0;

      if (requiredLineIndex < 0)
        return false;

      var line = lines[requiredLineIndex];
      var numberOfSentencesStr = line.Split(' ').Last();


      if (!long.TryParse(numberOfSentencesStr, out numberOfSentences))
        return false;

      return true;
    }

    public static bool GetAverageSentenceLength(string str, out double averageSentenceLength) {
      var lines = str.Split('\n');
      var requiredLineIndex = lines.Length - 1;
      averageSentenceLength = 0;

      if (requiredLineIndex < 0)
        return false;

      var line = lines[requiredLineIndex];
      var averageSentenceLengthStr = line.Split(' ').Last();

      if (!double.TryParse(averageSentenceLengthStr, out averageSentenceLength))
        return false;

      return true;
    }
  }
}
